--- Box2D plugin for STI
-- @module box2d
-- @author Landon Manning
-- @copyright 2015
-- @license MIT/X11


rotateVertex = (v, x, y, cos, sin, oy) ->
  oy = oy or 0

  vertex = {
    x: v.x,
    y: v.y - oy,
  }

  vertex.x = vertex.x - x
  vertex.y = vertex.y - y

  vx = cos * vertex.x - sin * vertex.y
  vy = sin * vertex.x + cos * vertex.y

  return vx + x, vy + y + oy


return {
	box2d_LICENSE     = "MIT/X11",
	box2d_URL         = "https://github.com/karai17/Simple-Tiled-Implementation",
	box2d_VERSION     = "2.3.0.2",
	box2d_DESCRIPTION = "Box2D hooks for STI.",

	--- Initialize Box2D physics world.
	-- @param world The Box2D world to add objects to.
	-- @return nil
	box2d_init: (map, world) ->
		assert(love.physics, "To use the Box2D plugin, please enable the love.physics module.")

		body      = love.physics.newBody(world)
		collision = {
			body = body,
		}

		addObjectToWorld = (objshape, vertices, userdata, object) ->

			if objshape == "polyline"
				shape = love.physics.newChainShape(false, unpack(vertices))
			else
				shape = love.physics.newPolygonShape(unpack(vertices))

			fixture = love.physics.newFixture(body, shape)

			fixture:setUserData(userdata)

			if userdata.properties.sensor == "true"
				fixture:setSensor(true)

			obj = {
        object: object,
				shape: shape,
				fixture: fixture,
			}

			table.insert(collision, obj)

		getPolygonVertices = (object, tile, precalc) ->
      ox, oy = 0, 0

      if not precalc
        ox = object.x
        oy = object.y

			vertices = {}
			for _, vertex in ipairs(object.polygon)
				-- table.insert(vertices, vertex.x)
				-- table.insert(vertices, vertex.y)
        table.insert(vertices, tile.x + ox + vertex.x)
        table.insert(vertices, tile.y + oy + vertex.y)

			return vertices

		calculateObjectPosition = (object, tile) ->
			o = {
        shape: object.shape,
				x: object.x,
				y: object.y,
				w: object.width,
				h: object.height,
				polygon: object.polygon or object.polyline or object.ellipse or object.rectangle
			}

			userdata = {
        object: o,
				properties: object.properties
			}

			if o.shape == "rectangle"
				o.r       = object.rotation or 0
				cos = math.cos(math.rad(o.r))
				sin = math.sin(math.rad(o.r))
				oy  = 0

				if object.gid
					tileset = map.tilesets[map.tiles[object.gid].tileset]
					lid     = object.gid - tileset.firstgid
					tile    = {}

					-- This fixes a height issue
					 o.y = o.y + map.tiles[object.gid].offset.y
					 oy  = tileset.tileheight

					for _, t in ipairs(tileset.tiles) do
						if t.id == lid then
							tile = t
							break

					if tile.objectGroup
						for _, obj in ipairs(tile.objectGroup.objects) do
							-- Every object in the tile
							calculateObjectPosition(obj, object)

						return
					else
						o.w = map.tiles[object.gid].width
						o.h = map.tiles[object.gid].height

				o.polygon = {
					{ x:o.x,       y:o.y       },
					{ x:o.x + o.w, y:o.y       },
					{ x:o.x + o.w, y:o.y + o.h },
					{ x:o.x,       y:o.y + o.h },
				}

				for _, vertex in ipairs(o.polygon)
					vertex.x, vertex.y = rotateVertex(vertex, o.x, o.y, cos, sin, oy)

				vertices = getPolygonVertices(o, tile, true)
				addObjectToWorld(o.shape, vertices, userdata, tile or object)
			elseif o.shape == "ellipse"
				if not o.polygon
					o.polygon = convertEllipseToPolygon(o.x, o.y, o.w, o.h)
				vertices  = getPolygonVertices(o, tile, true)
				triangles = love.math.triangulate(vertices)

				for _, triangle in ipairs(triangles)
					addObjectToWorld(o.shape, triangle, userdata, object)
			elseif o.shape == "polygon"
        precalc = false
        if not tile.gid then precalc = true
				vertices  = getPolygonVertices(o, tile, precalc)
				triangles = love.math.triangulate(vertices)

				for _, triangle in ipairs(triangles)
					addObjectToWorld(o.shape, triangle, userdata, object)

			elseif o.shape == "polyline"
        precalc = false
        if not tile.gid then precalc = true

				vertices	= getPolygonVertices(o, tile, precalc)
				addObjectToWorld(o.shape, vertices, userdata, object)

    -- FIXME: should instead be specifying the layer that causes collision higher up
		for _, layer in ipairs(map.layers)
      if layer.type == "tilelayer" then
        if layer.properties.collidable != "true"
          continue

        for _, tileset in ipairs(map.tilesets) do
            for _, tile in ipairs(tileset.tiles) do

                gid = tileset.firstgid + tile.id

                if tile.objectGroup and map.tileInstances[gid] then
                    for _, instance in ipairs(map.tileInstances[gid]) do
                      if instance.layer == layer then
                        for _, object in ipairs(tile.objectGroup.objects) do
                            -- Every object in every instance of a tile
                            calculateObjectPosition(object, instance)

		map.box2d_collision = collision
	,

	--- Draw Box2D physics world.
	-- @return nil
	box2d_draw = (map) ->
		collision = map.box2d_collision

		for _, obj in ipairs(collision)
			points = {collision.body:getWorldPoints(obj.shape:getPoints())}

			if #points == 4 then
				love.graphics.line(points)
			else
				love.graphics.polygon("line", points)
	,
}

--- Custom Properties in Tiled are used to tell this plugin what to do.
-- @table Properties
-- @field collidable set to "true", can be used on any Layer, Tile, or Object
-- @field sensor set to "true", can be used on any Tile or Object that is also collidable
